import torch
import numpy as np
from spacy.tokens import Doc
from typing import List, Dict, Tuple
from .filters_and_selectors import \
    load_perplex_scorer, \
    compute_sent_cosine_distance, \
    compute_delta_perplexity, \
    compute_sent_perplexity, \
    select_diverse_perturbations, \
    select_surprise_explanations, PredFormat
from .generations import \
    get_prompts, \
    load_generator, \
    get_random_idxes, \
    create_blanked_sents, \
    generate_on_prompts, \
    RANDOM_CTRL_CODES, ALL_CTRL_CODES
from .compute_perturbs import compute_edit_ops, SentenceMetadata
from .compute_perturbs.compute_edit_ops import compute_edit_ops

from .helpers import create_processor


import logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)  # pylint: disable=invalid-name



class Polyjuice(object):
    def __init__(self, model_path: str="uw-hai/polyjuice", is_cuda: bool=True) -> None:
        """The wrapper.

        Args:
            model_path (str, optional): The path to the generator.
                Defaults to "uw-hai/polyjuice". 
                No need to change this, unless you retrained a model
            is_cuda (bool, optional): whether to use cuda. Defaults to True.
        """
        # generator
        self.generator = None
        # validators
        self.polyjuice_generator_path = model_path
        self.perplex_scorer = None
        self.distance_scorer = None
        self.generator = None
        self.spacy_processor = None

        self.is_cuda = is_cuda and torch.cuda.is_available()

    def validate_and_load_model(self, model_name: str) -> bool:
        """Validate whether the generator or scorer are loaded.
        If not, load the model.

        Args:
            model_name (str): the identifier of the loaded part.
                Should be [generator, perplex_scorer].

        Returns:
            bool: If the model is successfully load.
        """
        if getattr(self, model_name, None):
            return True
        else:
            loader = getattr(self, f"_load_{model_name}", None)
            return loader and loader()

    def _load_perplex_scorer(self):
        logger.info("Setup perplexity scorer.")
        self.perplex_scorer = load_perplex_scorer(is_cuda=self.is_cuda)
        return True
    def _load_generator(self):
        logger.info("Setup Polyjuice.")
        self.generator = load_generator(
            self.polyjuice_generator_path, is_cuda=self.is_cuda)
        return True
    def _load_spacy_processor(self, is_space_tokenizer: bool=False):
        logger.info("Setup SpaCy processor.")
        self.spacy_processor = create_processor(is_space_tokenizer)
        return True

    ##############################################
    # validation
    ##############################################
    def _process(self, sentence: str):
        if not self.validate_and_load_model("spacy_processor"): return None
        return self.spacy_processor(str(sentence))

    def _compute_sent_cosine_distance(self, s1: str, s2: str):
        if not self.validate_and_load_model("distance_scorer"): return 1
        return compute_sent_cosine_distance(s1, s2, self.distance_scorer, is_cuda=self.is_cuda)

    def _compute_delta_perplexity(self, eops):
        if not self.validate_and_load_model("perplex_scorer"): return None
        return compute_delta_perplexity(eops, self.perplex_scorer, is_cuda=self.is_cuda)

    def _compute_sent_perplexity(self, sentence: str):
        if not self.validate_and_load_model("perplex_scorer"): return None
        return compute_sent_perplexity([sentence], self.perplex_scorer, is_cuda=self.is_cuda)[0]
    
    ##############################################
    # apis
    ##############################################

    def detect_ctrl_code(self, 
        orig: Tuple[str, Doc], 
        perturb: Tuple[str, Doc],
        eops: List=None) -> str:
        """Detect the perturbation type.

        Args:
            orig (Tuple[str, Doc]): The perturb-from sentence,
                either in str or SpaCy doc.
            perturb (Tuple[str, Doc]): The perturb-to sentence,
                either in str or SpaCy doc.
            eops (List, Optional): The editing operations, output of
                `compute_edit_ops` in 
                `polyjuice.compute_perturbs.compute_edit_ops`.
                If None, will be computed.
        Returns:
            str: The extracted code. If cannot be identified, return None.
        """
        orig = self._process(orig) if type(orig) == str else orig
        perturb = self._process(perturb) if type(perturb) == str else perturb
        if orig.text == perturb.text:
            return "equal"
        if eops is None:
            eops = compute_edit_ops(orig, perturb)
        meta = SentenceMetadata(eops)
        meta.compute_metadata(
            sentence_similarity=self._compute_sent_cosine_distance)
        if meta.primary and meta.primary.tag in ALL_CTRL_CODES:
            return meta.primary.tag
        else:
            return None

    def get_random_blanked_sentences(self, 
        sentence: Tuple[str, Doc], 
        pre_selected_idxes: List[int]=None,
        deps: List[str]=None,
        is_token_only: bool=False,
        max_blank_sent_count: int=3,
        max_blank_block: int=1) -> List[str]:
        """Generate some random blanks for a given sentence

        Args:
            sentence (Tuple[str, Doc]): The sentence to be blanked,
                either in str or SpaCy doc.
            pre_selected_idxes (List[int], optional): 
                If set, only allow blanking a preset range of token indexes. 
                Defaults to None.
            deps (List[str], optional): 
                If set, only select from a subset of dep tags. Defaults to None.
            is_token_only (bool, optional):
                blank sub-spans or just single tokens. Defaults to False.
            max_blank_sent_count (int, optional): 
                maximum number of different blanked sentences. Defaults to 3.
            max_blank_block (int, optional): 
                maximum number of blanks per returned sentence. Defaults to 1.

        Returns:
            List[str]: blanked sentences
        """
        if type(sentence) == str:
            sentence = self._process(sentence)
        indexes = get_random_idxes(
            sentence, 
            pre_selected_idxes=pre_selected_idxes,
            deps=deps,
            is_token_only=is_token_only,
            max_count=max_blank_sent_count,
            max_blank_block=max_blank_block
        )
        blanked_sents = create_blanked_sents(sentence, indexes)
        return blanked_sents

    def perturb(self, 
        orig_sent: Tuple[str, Doc], 
        blanked_sent: Tuple[str, List[str]]=None,
        is_complete_blank: bool=False, 
        ctrl_code: Tuple[str, List[str]]=None, 
        perplex_thred: int=10,
        num_perturbations: int=3,
        verbose: bool=False, 
        #is_include_metadata: bool=True,
        **kwargs) -> List[str]:
        """The primary perturbation function. Running example:
        Original sentence: 
            "It is great for kids."

        Args:
            orig_sent (Tuple[str, Doc]): 
                Original sentence, either in the form of str or SpaCy Doc.
            blanked_sents (Tuple[str, List[str]], optional): 
                sentences that contain blanks, e.g., "It is [BLANK] for kids."
                Defaults to None. If is None, the blank will be automatically placed.
                If is "" or incomplete form like "It is", set `is_complete_blank` to
                True below to allow the model to generate where to blank.
            is_complete_blank (bool, optional): 
                Whether the blanked sentence is already complete or not. 
                Defaults to False.
            ctrl_codes (Tuple[str, List[str]], optional): 
                The ctrl code (can be a list). Defaults to None. 
                If is None, will automatically become [resemantic, lexical,
                negation, insert, delete]. 
            perplex_thred (int, optional): 
                Perplexity filter for fluent perturbations. 
                we score both x and x' with GPT-2, and filter x' when the 
                log-probability (on the full sentence or the perturbed chunks) 
                decreases more than {perplex_thred} points relative to x
                Defaults to 5. If None, will skip filter.
            num_perturbations: 
                Num of max perturbations to collect. Defaults to 3.
            is_include_metadata: 
                Whether to return text, or also include other metadata and perplex score.
                Default to True.
            **kwargs: 
                The function can also take arguments for huggingface generators, 
                like top_p, num_beams, etc.
        Returns:
            List[str]: The perturbations.
        """
        logging.getLogger("transformers.generation_utils").setLevel(logging.ERROR)
        if not self.validate_and_load_model("generator"): return []
        orig_doc = self._process(orig_sent) if type(orig_sent) == str else orig_sent
        if blanked_sent:
            blanked_sents = [blanked_sent] if type(blanked_sent) == str else blanked_sent
        else:
            blanked_sents = self.get_random_blanked_sentences(orig_doc.text)
        if ctrl_code:
            ctrl_codes = [ctrl_code] if type(ctrl_code) == str else ctrl_code
            if not set(ctrl_codes).issubset(ALL_CTRL_CODES):
                logger.error(f"{set(ctrl_codes)-ALL_CTRL_CODES} is not a valid ctrl code. Please choose from {ALL_CTRL_CODES}.")
            is_filter_code = True
        else:
            ctrl_codes = RANDOM_CTRL_CODES
            is_filter_code = False
        prompts = get_prompts(
            doc=orig_doc,
            ctrl_codes=ctrl_codes, 
            blanked_sents=blanked_sents, 
            is_complete_blank=is_complete_blank)
        if verbose:
            logger.info("Generating on these prompts:")
            for p in prompts: logger.info(f" | {p}")
        generated = generate_on_prompts(
            generator=self.generator, prompts=prompts, **kwargs)
        merged = list(np.concatenate(generated))

        validated_set = []
        for _, generated in merged:
            # skip 
            if generated in validated_set or generated.lower() == orig_doc.text.lower(): 
                continue
            is_vaild = True
            generated_doc = self._process(generated)
            eop = compute_edit_ops(orig_doc, generated_doc)
            if perplex_thred is not None:
                pp = self._compute_delta_perplexity(eop)
                is_vaild = pp.pr_sent < perplex_thred and pp.pr_phrase < perplex_thred
            if is_vaild and is_filter_code:
                ctrl = self.detect_ctrl_code(orig_doc, generated_doc, eop)
                is_vaild = is_vaild and ctrl is not None and ctrl in ctrl_codes
            if is_vaild:
                validated_set.append(generated)
                #validated_changes.append(Munch(
                #    text=generated, doc=generated_doc, meta=meta, perplex=pp))
        if num_perturbations is None:
            num_perturbations = 1000
        sampled = np.random.choice(validated_set,
            min(num_perturbations, len(validated_set)), replace=False)
        return [str(s) for s in sampled]


    def select_diverse_perturbations(self, 
        orig_and_perturb_pairs: List[Tuple[str, str]],
        nsamples: int)-> List[Tuple[str, str]]:
        """Having each perturbation be represented by its token changes, 
        control code, and dependency tree strcuture, we greedily select the 
        ones that are least similar to those already selected. This tries 
        to avoid redundancy in common perturbations such as black -> white.

        Args: 
            orig_and_perturb_pairs (List[Tuple[str, str]]):
                A list of (orig, perturb) text pair. Not necessarily the 
                all the same orig text.
            nsamples (int): Number of samples to select

        Returns:
            List[Tuple[str, str]]: A subsample of (orig, perturb) pairs.
        """

        return select_diverse_perturbations(
            orig_and_perturb_pairs=orig_and_perturb_pairs,
            nsamples=nsamples,
            compute_sent_cosine_distance=self._compute_sent_cosine_distance,
            process=self._process)
        
    def select_surprise_explanations(self,
        orig_text: str,
        perturb_texts: List[str],
        orig_pred: PredFormat, 
        perturb_preds: List[PredFormat], 
        feature_importance_dict: Dict[str, float],
        agg_feature_importance_thred: float=0.1) -> List[Dict]:
        """Select surprising perturbations based on a feature importance map. 
        we estimate the expected change in prediction with feature attributions, 
        and select counterfactuals that violate these expectations, i.e., examples 
        where the real change in prediction is large even though importance scores 
        are low, and examples where the change is small but importance scores are high.

        Args:
            orig_doc (Doc): The original text.
            perturb_docs (List[Doc]): The perturbed texts.
            orig_pred (PredFormat): Prediction of the original text, in the format of 
                [{label: str, score: float}]
            perturb_preds (List[PredFormat]): Prediction of the perturbed texts,
                Each entry in the format of [{label: str, score: float}]
            feature_importance_dict (Dict[str, float]): The precomputed feature map.
            agg_feature_importance_thred (float, Optional): A threshold for classifying
                surprises.
        Returns:
            List[Dict]: Return the selected surprises in the format of:
            {
                case, [str]: Suprise filp | Suprise unflip, 
                pred, [str]: the prediction for the perturbation,
                changed_features, List[str]: the changed features
                perturb_doc: the perturbed doc))
            }
        """
        return select_surprise_explanations(
            orig_text=orig_text,
            perturb_texts=perturb_texts,
            orig_pred=orig_pred,
            perturb_preds=perturb_preds,
            feature_importance_dict=feature_importance_dict,
            processor=self._process,
            agg_feature_importance_thred=agg_feature_importance_thred
        )


PERETURB_TOK = "<|perturb|>"
BLANK_TOK = "[BLANK]"
SEP_TOK = "[SEP]"
EMPTY_TOK = "[EMPTY]"
ANSWER_TOK = "[ANSWER]"

# control codes
NEGATION = "negation"
QUANTIFIER = "quantifier"
SHUFFLE = "shuffle"
LEXCICAL = "lexical"
RESEMANTIC = "resemantic"
INSERT = "insert"
DELETE = "delete"
RESTRUCTURE = "restructure"

RANDOM_CTRL_CODES = [
    LEXCICAL, RESEMANTIC, NEGATION, INSERT, DELETE
]
ALL_CTRL_CODES = set([
    LEXCICAL, RESEMANTIC, NEGATION, INSERT, 
    DELETE, QUANTIFIER, RESTRUCTURE, SHUFFLE
])


import math
import numpy as np
from munch import Munch
from transformers import GPT2LMHeadModel, GPT2TokenizerFast

import torch
from copy import deepcopy

#########################################################################
### compute perplexity
#########################################################################

def _add_special_tokens(text, tokenizer):
    return tokenizer.bos_token + text + tokenizer.eos_token

def _tokens_log_prob(texts, model, tokenizer, batch_size=128, is_cuda=True):
        outputs = []
        for i in range(0, len(texts), batch_size):
            batch = texts[i : i + batch_size]
            outputs.extend(_tokens_log_prob_for_batch(batch, model, tokenizer, is_cuda=is_cuda))
        return outputs
    
def _tokens_log_prob_for_batch(texts, model, tokenizer, is_cuda=True):
    device = "cuda" if is_cuda else "cpu"
    outputs = []
    texts = [_add_special_tokens(text, tokenizer) for text in deepcopy(texts)]
    #encoding = tokenizer.batch_encode_plus(texts, return_tensors='pt')
    encoding = tokenizer.batch_encode_plus(texts, return_tensors='pt', truncation=True, padding=True)
    with torch.no_grad():
        ids = encoding["input_ids"].to(device)
        attention_mask = encoding["attention_mask"].to(device)
        #nopad_mask = ids != tokenizer.pad_token_id
        nopad_mask = ids != tokenizer.pad_token_id
        logits = model(ids, attention_mask=attention_mask)[0]
    
    for sent_index in range(len(texts)):
        sent_nopad_mask = nopad_mask[sent_index]
        sent_tokens = [tok 
            for i, tok in enumerate(encoding.tokens(sent_index))
            if sent_nopad_mask[i] and i != 0]
        sent_ids = ids[sent_index, sent_nopad_mask][1:]
        sent_logits = logits[sent_index, sent_nopad_mask][:-1, :]
        sent_logits[:, tokenizer.pad_token_id] = float("-inf")
        sent_ids_scores = sent_logits.gather(1, sent_ids.unsqueeze(1)).squeeze(1)
        sent_log_probs = sent_ids_scores - sent_logits.logsumexp(1)
        #sent_log_probs = cast(torch.DoubleTensor, sent_log_probs)
        #sent_ids = cast(torch.LongTensor, sent_ids)

        output = (sent_log_probs.cpu().numpy(), sent_ids.cpu().numpy(), sent_tokens)
        outputs.append(output)
    return outputs

def load_perplex_scorer(model_id = 'gpt2', is_cuda=True):
    model = GPT2LMHeadModel.from_pretrained(model_id)
    tokenizer = GPT2TokenizerFast.from_pretrained(model_id, use_fast=True, add_special_tokens=False)
    device = "cuda" if is_cuda else "cpu"
    tokenizer.add_special_tokens({"additional_special_tokens": ["<|pad|>"]})
    tokenizer.pad_token = "<|pad|>"
    model.resize_token_embeddings(len(tokenizer))
    model.eval()
    model.to(device)
    return Munch(model=model, tokenizer=tokenizer)
    
def reduce_perplex_prob(log_probs, log=False, reduce="prod"):
    tlen = log_probs.shape[0]
    if reduce == "prod":
        score = log_probs.sum()
    elif reduce == "mean":
        score = log_probs.logsumexp(0) - math.log(tlen)
    elif reduce == "gmean":
        score = log_probs.mean(0)
    elif reduce == "hmean":
        score = log_probs.neg().logsumexp(0).neg() + math.log(tlen)
    else:
        raise ValueError("Unrecognized scoring strategy: %s" % reduce)
    if not log:
        score = score.exp()
    return score.item()

def normalize_score(log_score, slen, alpha=0.8):
    #Elephant in the Room: An Evaluation Framework for Assessing Adversarial Examples in NLP
    return log_score/math.pow((5+slen)/6, alpha)

def compute_sent_perplexity(
    sentences, perplex_scorer, log=True, reduce="prod", is_normalize=False, is_cuda=True):
    """Compute the sentence perplexity. For filtering.

    Args:
        sentences ([type]): [description]
        perplex_scorer ([type]): [description]
        log (bool, optional): [description]. Defaults to True.
        reduce (str, optional): [description]. Defaults to "prod".
        is_normalize (bool, optional): [description]. Defaults to False.

    Returns:
        [type]: [description]
    """
    scores = []
    model, tokenizer = perplex_scorer.model, perplex_scorer.tokenizer
    outputs = _tokens_log_prob(sentences, model, tokenizer, is_cuda=is_cuda)
    for sent_log_prob, sent_ids, sent_tokens in outputs:
        score = reduce_perplex_prob(sent_log_prob, reduce=reduce, log=log)
        if is_normalize:
            score = normalize_score(score, len(sent_tokens))
        scores.append(score)
    return scores

def filter_by_sent_perplexity(sentences, perplex_scorer, thred=20, is_cuda=True):
    scores = compute_sent_perplexity(
        sentences, perplex_scorer, log=True, reduce="prod", is_normalize=False, is_cuda=is_cuda)
    idxes = np.where(np.array(scores) <= thred)[0]
    filtered =  [sentences[i] for i in idxes]

def compute_phrase_perplexity(
    sentence_phrase_tuples, perplex_scorer, 
    log=True, reduce="prod", is_normalize=False, is_cuda=True):
    scores = []
    sentence_phrase_tuples = sentence_phrase_tuples if type(sentence_phrase_tuples) != tuple else [sentence_phrase_tuples]
    if len(sentence_phrase_tuples) == 0:
        return scores
    model, tokenizer = perplex_scorer.model, perplex_scorer.tokenizer
    outputs = _tokens_log_prob([s[0] for s in sentence_phrase_tuples], model, tokenizer, is_cuda=is_cuda)
    for idx, (sentence, phrase) in enumerate(sentence_phrase_tuples):
        log_probs_all = outputs[idx][0]
        full_len = len(outputs[idx][1]) - 1
        if phrase:
            prefix_len = len(tokenizer(sentence.split(phrase)[0].strip())["input_ids"])
        else:
            prefix_len = 0
        phrase_len = len(tokenizer(phrase)["input_ids"])
        prefix_idx, phrase_idx = [0, prefix_len], [prefix_len, prefix_len+phrase_len]
        
        log_probs = log_probs_all[phrase_idx[0]:phrase_idx[1]]
        #print(sentence.split(phrase)[0].strip(), perplex_scorer.tokenizer(sentence.split(phrase)[0].strip()))
        #print(sentence, phrase, phrase_idx)
        full_sent_score = reduce_perplex_prob(log_probs_all, log=log, reduce=reduce)
        phrase_score = reduce_perplex_prob(log_probs, log=log, reduce=reduce)
        if is_normalize:
            full_sent_score = normalize_score(full_sent_score, full_len)
            phrase_score = normalize_score(phrase_score, phrase_len)
        scores.append((full_sent_score, phrase_score))
    return scores

def compute_delta_perplexity(edit_ops, perplex_scorer, is_normalize=False, is_cuda=True):
    """This is to compute the perplexity 

    Args:
        edit_ops ([type]): [description]
        perplex_scorer ([type]): [description]
        is_normalize (bool, optional): [description]. Defaults to False.

    Returns:
        [type]: [description]
    """
    tuples = []
    #print(metadata.primary.acore.doc.text)
    #print(metadata.primary.bcore.doc.text)
    edit_ops = [o for o in edit_ops if o.op != "equal"]
    for op in edit_ops:
        aphrase, bphrase = (op.fromz_full, op.toz_full) if \
            op.op == "insert" or op.op == "delete" else (op.fromz_core, op.toz_core)
        asent, bsent = aphrase.doc, bphrase.doc
        tuples += [(asent.text, aphrase.text), (bsent.text, bphrase.text)]
    #print(tuples)
    scores = compute_phrase_perplexity(tuples, perplex_scorer, 
        is_normalize=is_normalize, is_cuda=is_cuda)
    #print(scores)
    paired_scores = []
    for i in range(len(edit_ops)):
        # because of negative, it's i - i+1; lower the better.
        #print(scores[2*i])
        #print(scores[2*i+1])
        paired_scores.append(Munch(
            pr_sent=scores[2*i][0]-scores[2*i+1][0], 
            pr_phrase=scores[2*i][1]-scores[2*i+1][1]))
    paired_scores = sorted(paired_scores, key=lambda x: (
        max(x.pr_sent, x.pr_phrase)), reverse=True) # use the most ungrammar part as the 
    return paired_scores[0]



from nltk.translate.bleu_score import sentence_bleu
from zss import simple_distance, Node
import numpy as np
import edit_distance
from munch import Munch
import itertools

def compute_lev_distance(doc1, doc2):
    sm = edit_distance.SequenceMatcher(
        a=[t.text for t in doc1], b=[t.text for t in doc2])
    return 1-sm.ratio()

def add_node_to_root(root):
    if not root or not list(root.children):
        return None
    curr_node = Node(root.text.lower())
    for c in root.children:
        cnode = add_node_to_root(c)
        if cnode:
            curr_node.addkid(cnode)
    return curr_node
        
def get_nodes(doc):
    span = list(doc.sents)[0]
    return add_node_to_root(span.root)

def compute_tree_edit_distance(doc1, doc2):
    try:
        dist = simple_distance(get_nodes(doc1), get_nodes(doc2))
        return dist #/ len(doc1)
    except: return 0

def compute_closeness(docs, base_doc, sentence_similarity=None):
    sem_dist, tree_dist, edit_dist = [], [], []
    for doc in docs:
        if sentence_similarity:
            sem_dist.append(sentence_similarity(base_doc.text, doc.text))
        tree_dist.append(compute_tree_edit_distance(base_doc, doc))
        edit_dist.append(compute_lev_distance(base_doc, doc))
    return Munch(
        sem_dist=np.mean(sem_dist), 
        tree_dist=np.mean(tree_dist),
        edit_dist=np.mean(edit_dist), 
    )

def compute_self_bleu(docs, base_doc, kwargs=None):
    # if should just be augments around one example.
    scores = []
    if len(docs) == 0:
        return Munch(bleu4=1)
    included = []
    data_points = []
    for doc in docs:
        included.append(doc.text)
        data_points.append([d.text for d in doc])
    included.append(base_doc)
    data_points.append([d.text for d in base_doc])

    points = list(itertools.combinations(range(len(included)), 2))
    for i, j in points:
        scores.append(sentence_bleu([data_points[i]], data_points[j] ))
    return Munch(bleu4=np.mean(scores))


from .create_blanks import create_blanked_sents, get_random_idxes
from .generator_helpers import load_generator, generate_on_prompts, get_prompts
from .special_tokens import ALL_CTRL_CODES, RANDOM_CTRL_CODES


from ..helpers import _normalize_span, unify_tags


from spacy.tokens import Token, Span
from pattern.en import wordnet, number, numerals
import numpy as np
import itertools

####################################################################################
                ##### HELPER FUNCTIONS FOR META ########
####################################################################################
DEFAULT_NUM = 123456
NOUN_VERIATIONS = ["NOUN", "PROPN", "PRON"]

def all_synsets(word, pos=None):
    map = {
        'NOUN': wordnet.NOUN,
        'VERB': wordnet.VERB,
        'ADJ': wordnet.ADJECTIVE,
        'ADV': wordnet.ADVERB
        }
    if pos is None:
        pos_list = [wordnet.VERB, wordnet.ADJECTIVE, wordnet.NOUN, wordnet.ADVERB]
    else:
        pos_list = [map[pos]]
    ret = []
    for pos in pos_list:
        ret.extend(wordnet.synsets(word, pos=pos))
    return ret

def clean_senses(synsets):
    return [x for x in set(synsets) if '_' not in x]
def all_possible_synonyms(word, pos=None):
    ret = []
    for syn in all_synsets(word, pos=pos):
        # if syn.synonyms[0] != word:
        #     continue
        ret.extend(syn.senses)
    return clean_senses(ret)

def all_possible_antonyms(word, pos=None):
    ret = []
    for syn in all_synsets(word, pos=pos):
        if not syn.antonym:
            continue
        for s in syn.antonym:
            ret.extend(s.senses)
    return clean_senses(ret)

def all_possible_hypernyms(word, pos=None, depth=None):
    ret = []
    for syn in all_synsets(word, pos=pos):
        ret.extend([y for x in syn.hypernyms(recursive=True, depth=depth) for y in x.senses])
    return clean_senses(ret)
def all_possible_hyponyms(word, pos=None, depth=None):
    ret = []
    for syn in all_synsets(word, pos=pos):
        ret.extend([y for x in syn.hyponyms(recursive=True, depth=depth) for y in x.senses])
    return clean_senses(ret)
def all_possible_related(words, pos=None, depth=1):
    all_syns = [y for word in words for y in all_synsets(word, pos=pos)]
    # all_syns = [all_synsets(x, pos=pos) for x in words]
    # all_syns = [x[0] for x in all_syns if x]
    # return all_syns
    # print(all_syns)
    all_ancestors = [wordnet.ancestor(s1, s2) for s1, s2 in itertools.combinations(all_syns, 2)]
    all_ancestors = [x for x in all_ancestors if x]
    # print(all_ancestors)
    mapz = {x.lexname: x for x in all_ancestors}
    all_ancestors = list(mapz.values())
    all_descendents = [y for x in all_ancestors for y in x.hyponyms(recursive=True, depth=depth)]
    ret = [y for x in all_descendents for y in x.senses]
    return clean_senses(ret)


def is_subphrase(subset, superset):
    return subset.start >= superset.start and subset.end <= superset.end

def is_child_phrase(subset, superset):
    #print("is_child_phrase", subset.root.head in superset, subset.start >= superset.start and subset.end <= superset.end)
    if subset.start >= superset.start and subset.end <= superset.end:
        return True
    if all([t.head in superset for t in get_top_tokens(subset)]):
        return True
    return False #

def is_noun(span):
    normalized_span = _normalize_span(span)
    if not normalized_span: return False
    if type(normalized_span) == Token: return normalized_span.pos_ in NOUN_VERIATIONS 
    else: return normalized_span.root.pos_ in NOUN_VERIATIONS
def get_wordnet_info (func, span):
    normalized_span = _normalize_span(span)
    return set(func(normalized_span.lemma_)).union(set(func(normalized_span.text.lower())))
def get_hypernyms(root):
    if not is_noun(root): return set()
    hypers = get_wordnet_info(all_possible_hypernyms, root)
    if root.is_stop: return set()
    if root.pos_ in ["NOUN", "PROPN"]: hypers.add("something")
    if root.pos_ in ["PRON", "PROPN"] or root.ent_type_ == "PERSON": hypers.add("someone")
    if root.ent_type_: hypers.add(root.ent_type_)
    return hypers

def is_passive(span):
    if not span: return False
    children_set = set()
    """
    for t in span:
        children_set.add(t.dep_)
        print(t, list(t.children))
        children_set.union([i.dep_ for i in t.children])
    print(span, children_set)
    """
    #print([t.dep_ for t in span])
    return "auxpass" in set([t.dep_ for t in span])#, span.root.lemma_
    #return passive and passive.issubset(set(["agent", "auxpass"]))

def get_passive_token(span, only_in_span=False):
    if not span: return None
    children_set = set()
    """
    for t in span:
        children_set.add(t.dep_)
        print(t, list(t.children))
        children_set.union([i.dep_ for i in t.children])
    print(span, children_set)
    """
    head = span.root
    if not only_in_span:
        for t in span:
            if t.dep_ == "auxpass": 
                head = t.head
                break
        #while (head.pos_ not in ["VERB", "AUX"] or head.dep_.endswith("comp")) and head != head.head: head = head.head
    #print("get_passive_token", span, head)
    if head.pos_ not in ["VERB", "AUX"]: #or head.dep_.endswith("comp"):
        return None
    else:
        return head.lemma_

def get_negations(span, is_return_token=False):
    general_negations = set([
        "rare", "not", "no", "nothing", "unlike", "unless", "nobody", 
        "never", "seem", "nothing", "neither", "nowhere", "hardly", "scarcely",
        "barely", ])
    negation_phrases = set([
        "rather than", "suppose to", "other than", "don't", "doesn't", "didn't"
        "use to", "should be", "can be",  "hard to", "isn't", "shouldn't"
    ])
    
    general_sets, head_sets = set(), set()
    if not span: general_sets, head_sets

    def add_negations(t):
        general_sets.add(t if is_return_token else t.lemma_)
    for t in span:
        if t.lemma_ in general_negations or t.dep_ == "neg":
            add_negations(t)
        elif "n't" in t.text.lower():
            add_negations(t)
        else:
            head_sets.add(t.lemma_)
    for n in negation_phrases:
        if n in span.lemma_:
            add_negations(span)
            head_sets = set()
    return general_sets, head_sets

def is_subjunctive(span):
    text = span.text.lower().replace("'ve", "have")
    lemma = span.lemma_.replace("'ve", "have")
    if not span: return False
    #print (span.text.lower(), span.lemma_)
    return span and (
        "should have" in text or  \
    "could have" in text or "should have" in lemma or  "could have" in lemma)

def get_tense(span, only_in_span=False):
    if not span: return None, None
    # get the verb
    head = span.root
    if not only_in_span:
        while (head.pos_ not in ["VERB", "AUX"] or head.dep_.endswith("comp")) and head != head.head: head = head.head
    # childrens 
    if head.pos_ not in ["VERB", "AUX"] or head.dep_.endswith("comp"):
        return None, None
    #print(span, head, head.dep_, list(head.children))
    if "will" in [c.lemma_ for c in list(head.children)+[head]]: return "future", head.lemma_
    sub_verb = [c for c in head.children if c.lemma_ in ["can", "do"]]
    core = sub_verb[0] if sub_verb else head
    return "past" if core.tag_ in ["VBD", "VBN"] else "present", core.lemma_

def get_quantifiers(span):
    # determine if it's changing quantifier
    # case 1: words occur: all, every, some, no, nothing, something, at most, at least
    # case 2: is changing nummod/quantmod or the root of the change is NUM

    def get_number(r):
        r = r.lower()
        #print(r)
        if r == "zero" or str(r) == "0": return 0
        elif number(r) > 0: return number(r)
        #elif r == "both": return 2
        elif r == "dozen": return 12
        elif r in ["no", "nobody"]: return 0
        #elif r in ["all", "every", "any", "each"]: return 40404
        #elif r in ["few"]: return 50505
        #elif r in ["many", "much"]: return 20202
        #elif r in ["some", "several", "certain"]: return 30303
        #elif r in ["other", "another"]: return 10101
        return -1
    general_quantifiers = set([
        "all", "every", "some", "any", "several", "certain", "each", "both", "many","much",
        "no", "dozen", "lot", "few", "other", "another", "nobody", "most", "least"])
    if not span:
        return set(), set(), ""
    # only allow it if we know it's a modifier change
    tops = span #get_top_tokens(span)
    #if span.root.pos_ != "NUM" and all([unify_tags(s) != "modifier" for s in tops]):
    #    return set(), set()
    # and t.ent_type_ in ["PERCENT", "QUANTITY", "ORDINAL", "CARDINAL"]
    # and t.head.ent_type_ in ["PERCENT", "QUANTITY", "ORDINAL", "CARDINAL"]
    #print([(t.lemma_, t.pos_, t.dep_) for t in tops])
    related = set([t.lemma_ for t in tops if (t.lower_ in general_quantifiers and \
        (t.dep_ in ["det"] or unify_tags(t.dep_) in ["subj", "obj"])) or \
        (t.dep_ in ["nummod", "quantmod"]) or \
        (t.pos_ == "NUM" ) or \
        (t.head.pos_ == "NUM" and unify_tags(t.dep_) in ["modifier"])
        ])
    #print(tops)
    # also extract the numbers
    numbers = set([get_number(r) for r in related if get_number(r) >= 0])
    #print(related, numbers)
    #if "a" in related: numbers.add(1)
    #if "an" in related: numbers.add(1)
    if "both" in related: numbers.add(2)
    if "no" in related or "none" in related: numbers.add(0)
    if "dozen" in related: numbers.add(12)
    #print(span, related, numbers)
    related = set([r for r in related if get_number(r) == -1])
    return related.union(numbers), numbers, span.root.lemma_.lower()

def unify_pos(pos):
    #if pos in ["VERB", "AUX"]: return "VERB"
    if pos in NOUN_VERIATIONS: return "NOUN"
    elif pos.endswith("CONJ") or pos == "ADP": return "ADP"
    return pos

def get_top_tokens(span, is_filter_stop=False):
    normalized_span = _normalize_span(span)
    roots = []
    if not normalized_span: 
        return roots
    for token in normalized_span:
        if is_filter_stop and (token.is_stop or token.is_punct):
            continue
        if any([t.is_ancestor(token) for t in normalized_span]):
            continue
        roots.append(token)
    return roots

def is_single_change(span):
    include_tokens = [t for t in _normalize_span(span) if not t.dep_ in ["det", "prt"] and not t.is_punct]
    top_tokens = get_top_tokens(span)
    #print("is_single_change", include_tokens)
    #print("is_single_change", top_tokens)
    return len(include_tokens) <= 1 and len(top_tokens) <= 1

def is_same_structure(t1, t2):
    a, b = get_top_tokens(t1), get_top_tokens(t2)
    #print("is_same_structure", t1, t2)
    dep1 = set([unify_tags(r.dep_) for r in a])
    dep2 = set([unify_tags(r.dep_) for r in b])
    dep_intersect = dep1.intersection(dep2) - set(["aux"])
    pos_intersect = set([unify_pos(r.pos_) for r in a]).intersection(set([unify_pos(r.pos_) for r in b]))
    #print("is_same_structure", dep1, dep2)
    head1 = set([(unify_tags(r.head.dep_), unify_tags(r.head.pos_)) for r in a])
    head2 = set([(unify_tags(r.head.dep_), unify_tags(r.head.pos_)) for r in b])
    head_intersect = head1 == head2 #and len(set([unify_tags(r.head.dep_) for r in a])-set(["ROOT"])) > 0
    #print(dep_intersect, head_intersect)
    #print(a, [r.head for r in a], [unify_pos(r.pos_) for r in a], [r.dep_ for r in a])
    #print(b, [r.head for r in b], [unify_pos(r.pos_) for r in b], [r.dep_ for r in b])
    #print(head1)
    #print(dep_intersect)
    #print(pos_intersect)
    #print(head1, head_intersect)
    #print(t1, is_single_change(t1))
    #print(t2, is_single_change(t2))
    #if "ROOT" in dep_intersect and is_single_change(t1) and is_single_change(t2):
    #    return True
    #elif "ROOT" not in dep_intersect:
    return  len(dep_intersect) or head_intersect
    if "ROOT" not in dep2 and "ROOT" not in dep1:
        return  (len(pos_intersect) > 0 and head_intersect) or len(dep_intersect) > 0
    else:
        return  (len(dep_intersect) > 0 and len(pos_intersect) > 0) and head_intersect



class PhraseMetadata(object):
    def __init__(self, op, acore, bcore, afull, bfull):
        self.acore, self.bcore = acore, bcore
        self.afull, self.bfull = afull, bfull
        #self.acore, self.bcore = araw, braw
        
        self.op = op
        self.edit_dist = len(self.acore) + len(self.bcore)
        self.sem_dist = -1
        self._set_dep_and_pos()
        self.tag = None

    def _set_dep_and_pos(self):
        if self.op == "insert" or \
            (self.op != "delete" and (
                not _normalize_span(self.acore) or \
                _normalize_span(self.acore).root.dep_ == "det")):
            core = self.bcore
        else:
            core = self.acore
        core = _normalize_span(core)
        try:
            self.dep, self.pos = core.root.dep_, core.root.pos_
        except:
            self.dep, self.pos = "", ""

    def short_str(self):
        return f"[{self.op}] {self.acore} -> {self.bcore}"

    def __repr__(self):
        string = self.short_str()
        #string += f"\n\t== {self.afull}"
        #string += f"\n\t== {self.bfull}"
        #string += f"\n\tstructure remain: {self._is_structure_remain()}"
        #string += f"\n\tlocal: {self._is_local()}"
        string += f"\n\ttag: {self.tag}"
        string += f"\n\tdep: {unify_tags(self.dep)}"
        string += f"\n\tsem_dist: {self.sem_dist:.3f}"
        return string
    
    def _is_local(self):
        # TODO: change this part maybe?
        # How many different dep trees are changed
        a, b = (self.acore, self.bcore) if self.op == "replace" else (self.afull, self.bfull)
        return max(len(self.acore), len(self.bcore)) <= 5 and \
            (len(get_top_tokens(a)) == 1 and len(get_top_tokens(b)) == 1)

    def _is_structure_remain(self):
        t1, t2 = (self.acore, self.bcore) if self.op == "replace" else (self.afull,self.bfull)
        return is_same_structure(t1, t2)

    def compute_sem_dist(self, sentence_similarity, reset=False):
        #print(self.afull, self.bfull, sentence_similarity(self.afull.text, self.bfull.text))
        #print(self.acore, self.bcore, sentence_similarity(self.acore.text, self.bcore.text))
        if reset:
            self.sem_dist = -1
        if self.sem_dist == -1: 
            l1 = set([a.lemma_.lower() for a in self.acore])
            l2 = set([a.lemma_.lower() for a in self.bcore])
            stopset = set(["a", "an", "the", "this", "that", "be"])
            if (l1-l2).issubset(stopset) and (l2-l1).issubset(stopset):
                self.sem_dist = 0.01
            else:
                self.sem_dist = min(
                    sentence_similarity(self.afull.text, self.bfull.text),
                    sentence_similarity(self.acore.text, self.bcore.text),
                ) if sentence_similarity else -1

    def compute_metadata(self, sentence_similarity, is_overwrite_local=False):
        tag = None
        is_local, is_structure_remain = self._is_local(), self._is_structure_remain()
        #print("structure_remain", is_structure_remain)
        acore, bcore = _normalize_span(self.acore), _normalize_span(self.bcore)
        afull, bfull = _normalize_span(self.afull), _normalize_span(self.bfull)
        aroot = acore.root if acore else None
        broot = bcore.root if bcore else None

        if is_overwrite_local:
            is_local = True

        change_set = [aroot.text.lower() if aroot else "", broot.text.lower() if broot else ""]
        def is_wordnet_change(aroot, broot):
            return not (aroot.is_stop and broot.is_stop) and aroot.lemma_ not in ["have", "be", "get"] and broot.lemma_ not in ["have", "be"]
        xor = lambda func: (func(acore) and not func(bcore)) or (func(bcore) and not func(acore))
        ### negation
        negs1, neg_heads1 = get_negations(acore)
        negs2, neg_heads2 = get_negations(bcore)
        #print(negs1, negs2, neg_heads1, neg_heads2)
        if is_local and tag is None and (
            (negs1 != negs2) or \
            (self.op == "replace" and change_set == set(["and", "but"])) or \
            xor(is_subjunctive)):
            tag = "negation"
        ### quantifiers
        quants1, nums1, base1 = get_quantifiers(acore)
        quants2, nums2, base2 = get_quantifiers(bcore)
        #print(base1, base2)
        #print("get_quantifiers", quants1, nums1, base1)
        #print("get_quantifiers", quants2, nums2, base2)
        #print(quants1, quants2, nums1, nums2)
        # TODO: what is this one doing?
        if tag is None and is_local and (base1 == base2 or \
            self.op != "replace"  or \
            (self.op == "replace" and acore.root.head.lemma_ == bcore.root.head.lemma_)):
            if quants1 and quants2 and quants1 != quants2: #quants1 and quants2 and
                tag = "quantifier"
            n1, n2 = list(nums1-nums2)[0] if nums1-nums2 else DEFAULT_NUM, list(nums2-nums1)[0] if nums2-nums1 else DEFAULT_NUM
            if tag is None and ((n1 != DEFAULT_NUM) or (n2 != DEFAULT_NUM)):
                tag = "quantifier"
            if set([acore.text.lower(),bcore.text.lower()]).issubset(set(["a", "one", "an"])):
                tag = None
        ### split the nouns
        # TODO: should this be just for NOUNs

        def get_span_info(func, span, full_span): 
            return func(span) if len(span) and func(span) else func(full_span)
        if tag is None:
            p1 = get_span_info(get_passive_token, acore, afull)
            p2 = get_span_info(get_passive_token, bcore, bfull)
            #print(is_subjunctive(self.acore), is_subjunctive(self.bcore))
            tense1, v1 = get_span_info(get_tense, acore, afull)
            tense2, v2 = get_span_info(get_tense, bcore, bfull)
            if (xor(is_passive) and p1 and p1 == p2) or (tense1 != tense2 and v1 == v2):
                tag = "restructure"
        if tag is None and self.op == "replace" and is_local and is_structure_remain:
            hypers1, hypers2 = get_hypernyms(aroot), get_hypernyms(broot)
            #print(aroot, hypers1)
            #print(broot, hypers2)
            if is_wordnet_change(aroot, broot) and (acore.lemma_.lower() in hypers2 or acore.text.lower() in hypers2):
                tag = "lexical"
            elif is_wordnet_change(aroot, broot) and (bcore.lemma_.lower() in hypers1 or bcore.text.lower() in hypers1):
                tag = "lexical"
            else:
                syms = get_wordnet_info(all_possible_synonyms, acore)
                ayms = get_wordnet_info(all_possible_antonyms, acore)
                #if False and bcore.lemma_ in syms or bcore.text.lower() in syms: 
                #    tag = "paraphrase"
                #el
                if bcore.lemma_ in ayms or bcore.text.lower() in ayms:
                    tag = "lexical"
                else:
                    #self.compute_sem_dist(sentence_similarity)
                    #if False and self.sem_dist <= 0.1 and self.sem_dist >= 0:
                    #    tag = "paraphrase" 
                    #el
                    if is_single_change(acore) and is_single_change(bcore):
                        tag = "lexical"
                    else:
                        tag = "resemantic"

        if tag is None and self.op != "replace" and is_structure_remain:
            #self.compute_sem_dist(sentence_similarity)
            #if self.sem_dist >= 0 and self.sem_dist <= 0.1: tag = "paraphrase"
            #else: tag = "constraint"
            #self.compute_sem_dist(sentence_similarity)
            #tag = "paraphrase" if self.sem_dist <= 0.1 and self.sem_dist >= 0 else self.op
            tag = self.op # TODO: should we keep the semantic tag?
        if tag is None and is_local and not is_structure_remain:
            #self.compute_sem_dist(sentence_similarity)
            #tag = "paraphrase" if self.sem_dist <= 0.1 and self.sem_dist >= 0 else "restructure"
            tag = "other"
        if tag is None and not is_local: tag = "other"
        if tag is None: tag = "global"
        #print(tag, self.tag)
        self.tag = tag
        

class SentenceMetadata(object):
    def __init__(self, ops):
        self.phrases = [
            PhraseMetadata(op.op, 
                op.fromz_core, op.toz_core, 
                op.fromz_full, op.toz_full,
            ) for op in ops]
        self.primary = None
    def to_dict(self):
        return self.primary.to_dict() if self.primary else {}
    def export_sentence(self):
        if self.primary:
            return {
                "atext": self.primary.acore.doc.text,
                "btext":self.primary.bcore.doc.text
            }
        else:
            return {"atext": "", "btext": ""}
    def __repr__(self):
        if self.primary:
            primary = str(self.primary)
            others = [str(p) for p in self.phrases if str(p) != primary]
            strs = [primary] + others
        else:
            strs = [str(p) for p in self.phrases]
        string = "\n".join(strs)
        return string

    def short_str(self):
        string = "\n".join([p.short_str() for p in self.phrases])
        return string

    def _get_change_intersections(self):
        exclude_list = set(["be", "this", "there", "the", "a", "an", "that", "those", "-pron-"])
        pos_list = ["NOUN", "ADJ", "ADV","NUM", "PRON", "PROPN", "VERB"]
        achanges, bchanges = set(), set()
        #print(self.nonoverlap_phrases(use_full=False))
        phrases = self.nonoverlap_phrases(use_full=False)
        filter_word = lambda s: not s.is_punct and s.pos_ in pos_list and s.lemma_.lower() not in exclude_list
        for p in phrases: #self.phrases:
            cur_a = set([s.lemma_.lower() for s in p.acore if filter_word(s)])
            cur_b = set([s.lemma_.lower() for s in p.bcore if filter_word(s)])
            achanges.update(cur_a-cur_b)
            bchanges.update(cur_b-cur_a)
        intersect = achanges.intersection(bchanges)
        return achanges, bchanges, intersect, phrases
        
    def get_major_phrases(self):
        if not self.primary:
            return []
        primary = self.primary
        allowed_phrase_combinations = []
        if primary.tag in ["shuffle"]:
            new_phrases = []
            achanges, bchanges, intersect, phrases = self._get_change_intersections()
            for p in phrases:
                if any([i in p.acore.lemma_.lower() for i in intersect]):
                    bcore = p.bcore.doc[p.bcore.start:p.bcore.start]
                    np = PhraseMetadata("delete", p.acore, bcore, p.afull, p.bfull)
                    np.tag = "shuffle"
                    new_phrases.append(np)
                if any([i in p.bcore.lemma_.lower() for i in intersect]):
                    acore = p.acore.doc[p.acore.start:p.acore.start]
                    np = PhraseMetadata("insert", acore, p.bcore, p.afull, p.bfull)
                    np.tag = "shuffle"
                    new_phrases.append(np)
            allowed_phrase_combinations.append(new_phrases)
        elif primary.tag == "restructure":
            pass
        elif primary.tag in ["lexical", "resemantic"]:
            allowed_phrase_combinations.append([primary])
        else:
            phrases = self.nonoverlap_phrases(use_full=False)
            if len(phrases) > 1 or primary.tag not in ["lexical", "resemantic"]:
                allowed_phrase_combinations.append(phrases)
        return allowed_phrase_combinations

    def compute_metadata(self, sentence_similarity, is_select_primary=True):
        for p in self.phrases:
            p.compute_metadata(sentence_similarity)
            #print(p)
        if not is_select_primary or not self.phrases:
            return
        self.select_primary(sentence_similarity)
        
        #print("--- primary ---")
        #print(self.primary)
    
    def select_primary(self, sentence_similarity):
        def get_order(p, treat_insert_diff=False):
            order = {
                "global": -1,
                #"lexical":3,
                #"hyponym":3,
                #"wordnet":3,
                #"other": 3,
                #"quantifier": 0,
                #"constraint": 1,
                #"restructure": 2,
                "paraphrase": 1000,
                "restructure": 1,
            }
            if treat_insert_diff:
                order["insert"] = 500
                order["delete"] = 500
            return order[p.tag] if p.tag in order else 2
        if len(self.phrases) == 1: 
            self.primary = self.phrases[0]
            return
        for p in self.phrases:
            p.compute_sem_dist(sentence_similarity)
        
        achanges, bchanges, intersect, _ = self._get_change_intersections()
        #print(intersect, achanges, bchanges)
        # get phrases if they relate to the intersection
        has_shuffle = len(intersect) > 0 and (len(intersect) == len(achanges) or len(intersect) == len(bchanges)) \
            and "there" not in achanges and \
            "there" not in bchanges
        # if re-structuring, compare the depth of all the dependency trees and the children relationship
        # if some other core is within the restructure full, then prioritize the restructruing
        # rationale: they would be related to the restructuring.
        phrases_for_sort = []
        p_restructures = self.nonoverlap_phrases(use_full=False, phrases=[p for p in self.phrases if p.tag == "restructure"])

        for p in self.phrases:
            if p.tag == "restructure":
                continue
            if any([is_child_phrase(p.acore, pr.afull) and is_child_phrase(p.bcore, pr.bfull) for pr in p_restructures]):
                continue
            phrases_for_sort.append(p)
        #if phrases_for_sort:
        phrases_for_sort += p_restructures
        #else:
        #    phrases_for_sort = self.phrases
        phrases_for_sort = sorted(phrases_for_sort, key=lambda p: (get_order(p), -p.sem_dist))
        #print([(p, get_order(p)) for p in phrases_for_sort])
        #phrases_for_shuffle = sorted(phrases_for_sort, key=lambda p: (-p.sem_dist))
        #print(phrases_for_sort)
        if len(phrases_for_sort) == 0: 
            self.primary = None
            return
        #if len(intersect) / len(achanges) >= 0.5 and len(intersect) / len(bchanges) >= 0.5:
        #    print("shuffle")
        has_multiple = len(set([p.op for p in phrases_for_sort if p.op in ["insert", "delete"] and \
            p.tag in ["insert", "delete", "restructure"] ])) > 1
        #print(phrases_for_sort[0].op in ["insert", "delete"], has_shuffle, has_multiple)

        cur_a = set([s.lemma_.lower() for s in phrases_for_sort[0].acore if not s.is_stop and not s.is_punct])
        cur_b = set([s.lemma_.lower() for s in phrases_for_sort[0].bcore if not s.is_stop and not s.is_punct])

        if any([p.tag == "global" for p in phrases_for_sort]) or \
            len([p for p in phrases_for_sort if p.tag != "paraphrase"]) >= 5:
            self.primary = phrases_for_sort[0]
            self.primary.tag = "global"
        elif has_shuffle: #and intersect.intersection(cur_a.union(cur_b)) and all([p.tag != "restructure" for p in self.phrases]):
            #phrases_shuffle = [s for s in self.phrases if any([s.lemma_ in intersect for s in list(p.acore) + list(p.bcore)])]
            self.primary = phrases_for_sort[0]
            if all([p.tag != "restructure" for p in self.phrases]): 
                self.primary.tag = "shuffle"
        elif phrases_for_sort[0].op in ["insert", "delete"] and has_multiple:
            self.primary = phrases_for_sort[0]
            inserts = [p for p in phrases_for_sort if p.op == "insert"]
            deletes = [p for p in phrases_for_sort if p.op == "delete"]
            self.primary.tag = "resemantic"
            for i, d in itertools.product(inserts, deletes):
                if not is_same_structure(i.bcore, d.acore):
                    self.primary.tag = "global"
                    break
        else:
            self.primary = phrases_for_sort[0]
        if self.primary.tag == "other": self.primary.tag = "global"

    def nonoverlap_phrases(self, use_full, merge_entire=False, phrases=None):
        if not phrases:
            phrases = self.phrases
        a, b = ("afull", "bfull") if use_full else ("acore", "bcore")
        if merge_entire:

            acore_starts = [p.acore.start for p in phrases]
            acore_ends = [p.acore.end for p in phrases]
            bcore_starts = [p.bcore.start for p in phrases]
            bcore_ends = [p.bcore.end for p in phrases]

            afull_starts = [p.afull.start for p in phrases]
            afull_ends = [p.afull.end for p in phrases]
            bfull_starts = [p.bfull.start for p in phrases]
            bfull_ends = [p.bfull.end for p in phrases]

            adoc, bdoc = phrases[0].acore.doc, phrases[0].bcore.doc
            acore = adoc[min(acore_starts):max(acore_ends)]
            bcore = bdoc[min(bcore_starts):max(bcore_ends)]
            afull = adoc[min(afull_starts):max(afull_ends)]
            bfull = bdoc[min(bfull_starts):max(bfull_ends)]
            if not acore: op = "insert"
            elif not bcore: op = "delete"
            else: op = "replace"
            """
            print(pstart.acore)
            print(pstart.bcore)
            print(afull.acore)
            print(pstart.bcore)
            print(op, acore, bcore, afull, bfull, acore, bcore)
            """
            try:
                normalized_phrases = [PhraseMetadata(op, acore, bcore, afull, bfull)]
            except:
                normalized_phrases = []
        else:
            phrases = sorted(phrases, key=lambda l: (len(getattr(l, a)) , len(getattr(l, b)) ))
            normalized_phrases = []
            for idx, n in enumerate(phrases):
                is_subset = False
                for m in phrases[idx+1:]:
                    #print(n.short_str(), getattr(n, a).start, getattr(n, a).end, getattr(n, b).start, getattr(n, b).end)
                    #print(m.short_str(), getattr(m, a).start, getattr(m, a).end, getattr(m, b).start, getattr(m, b).end)
                    #if (getattr(n, a).start >= getattr(m, a).start and getattr(n, a).end <= getattr(m, a).end or n.op == "insert") and \
                    #    (getattr(n, b).start >= getattr(m, b).start and getattr(n, b).end <= getattr(m, b).end  or n.op == "delete"):
                    #print(getattr(n, a), getattr(m, a))
                    #print(getattr(n, b), getattr(m, b))
                    if is_subphrase(getattr(n, a), getattr(m, a)) and is_subphrase(getattr(n, b), getattr(m, b)):
                        is_subset = True
                        break
                    elif not (getattr(n, a).start > getattr(m, a).end or getattr(m, a).start > getattr(n, a).end):
                        return []
                if not is_subset:
                    normalized_phrases.append(n)
            normalized_phrases = sorted(normalized_phrases, key=lambda p: p.afull.start)
        return normalized_phrases


import difflib
from ..helpers import _normalize_chunk
from munch import Munch

#########################################################################
### helper functions for computing polyjuice changes
#########################################################################
def is_punct_phrase(idxes, doc):
    return idxes[1] - idxes[0] == 0 or len(_normalize_chunk(doc, idxes[0], idxes[1]))==0

def complete_tree_pair(aidx, bidx, adoc, bdoc, ops, adoc_range=None, bdoc_range=None):
    """this function extend the subphrases to their matching subtrees.

    Returns:    
        [Span, Span]: the extended aspan, and bspan
    """
    if is_punct_phrase(aidx, adoc) and is_punct_phrase(bidx, bdoc):
        return adoc[aidx[0]:aidx[1]], bdoc[bidx[0]:bidx[1]]
    for full_tree in [False, True]:
        aidx_extend = complete_one_tree(aidx, adoc, range_idxes=adoc_range, full_tree=full_tree)
        bidx_extend = complete_one_tree(bidx, bdoc, range_idxes=bdoc_range, full_tree=full_tree)
        # this part allows overflow
        _bidx_candidate = list(bidx_extend) + [match_idx_in_eops(idx, ops, "from", True) for idx in aidx_extend]
        bidx_new = [max(0, min(_bidx_candidate)), min(len(bdoc), max(_bidx_candidate))]        
        _aidx_candidate = list(aidx_extend) + [match_idx_in_eops(idx, ops, "to", True) for idx in bidx_new]
        aidx_new = [max(0, min(_aidx_candidate)), min(len(adoc), max(_aidx_candidate))]

        if not is_punct_phrase(aidx_new, adoc) and not is_punct_phrase(bidx_new, bdoc):
            break
    return adoc[aidx_new[0]:aidx_new[1]], bdoc[bidx_new[0]:bidx_new[1]]


def compute_edit_ops(adoc, bdoc, only_return_edit=False):
    stopwords = ["the", "an", "a", "be", ""]
    eops_raw = difflib.SequenceMatcher(
        a=[t.text for t in adoc],
        b=[t.text for t in bdoc]).get_opcodes()
    eops = []
    for op,f1,f2,t1,t2 in eops_raw:
        if only_return_edit and (
            op == "equal" or 
                (_normalize_chunk(adoc, f1, f2).lemma_.lower() in stopwords and \
                _normalize_chunk(bdoc, t1, t2).lemma_.lower() in stopwords)):
            continue
        s1, s2 = complete_tree_pair([f1, f2], [t1, t2], adoc, bdoc, eops_raw)
        #print(op,f1,f2,t1,t2)
        #print(s1)
        #print(s2)
        #print(adoc[f1:f2])
        #print(bdoc[2:2])
        eops.append(Munch(op=op,
            # all are spans
            fromz_core=adoc[f1:f2], toz_core=bdoc[t1:t2],
            fromz_full=s1, toz_full=s2,
        ))
    return eops

def complete_one_tree(span_idxes, doc, range_idxes=None, full_tree=True):
    """Compute the complete subtree for a given span

    Args:
        span_idxes ([int, int]): the indexes of the given span to be extended
        doc (Doc): The spacy doc
        range_idxes ([int, int]], optional): the maixmum indexs. Defaults to None.
        full_tree (bool, optional): if use the entire tree. Defaults to True.

    Returns:
        [int, int]: the extended tree
    """
    span_start, span_end = span_idxes
    if span_end - span_start == 0:
        return [span_start, span_end]
    if all([doc[i].is_punct for i in range(span_start, span_end)]):
        return [span_start, span_start]
    span = doc[span_start: span_end]
    if full_tree:
        left_edge, right_edge = range_idxes if range_idxes else [0, len(doc)]
        for token in span:
            if any([t.is_ancestor(token) for t in span]):
                continue
            head = token.head
            left_edge = max(left_edge, head.left_edge.i)
            right_edge = min(right_edge, head.right_edge.i)
    else:
        all_tokens = set()
        for token in span:
            head = token.head
            subtrees = set([head, token]) #(set(head.subtree) - set(token.subtree)) | set([token])
            all_tokens |= subtrees
        #left_edge = max(left_edge, curr_left_edge)
        #right_edge = min(right_edge, curr_right_edge)
        left_edge = min([a.i for a in all_tokens] + [span_start])
        right_edge = max([a.i for a in all_tokens] + [span_end-1])
    while left_edge < span_start and doc[left_edge].is_punct:
        left_edge += 1
    while right_edge > span_end and doc[right_edge].is_punct:
        right_edge -= 1
    left_idx, right_idx = left_edge, right_edge
    if left_idx <= right_idx: # or root == span.root.head:
        return [left_idx, right_idx+1]
    else:
        return [span_start, span_end]

def match_idx_in_eops(source_idx, ops, source_str="from", allow_overflow=False):
    sidx_start = 1 if source_str=="from" else 3
    tidx_start = 3 if source_str=="from" else 1
    sidx_end, tidx_end = sidx_start + 1, tidx_start + 1
    target_idx = -1
    if allow_overflow and source_idx == ops[-1][sidx_start+1]:
        return ops[-1][tidx_end]
    for op in ops:
        if (source_idx >= op[sidx_start] and source_idx < op[sidx_end]) or \
            (source_idx == op[sidx_start] and source_idx == op[sidx_end]):
            if op[tidx_start] == op[tidx_end]:
                target_idx = op[tidx_start]
            else:
                target_idx = source_idx - op[sidx_start] + op[tidx_start]
            if not allow_overflow and target_idx >= ops[-1][tidx_end]-1:
                target_idx = ops[-1][tidx_end]-1
            break
    return target_idx



from tqdm import tqdm
from transformers import pipeline, AutoTokenizer, AutoModelForCausalLM
import torch
import re
import itertools
from .special_tokens import BLANK_TOK, SEP_TOK, EMPTY_TOK, ANSWER_TOK, PERETURB_TOK
from .create_blanks import create_blanked_sents

def load_generator(model_path="uw-hai/polyjuice", is_cuda=True):
    return pipeline("text-generation", 
        model=AutoModelForCausalLM.from_pretrained(model_path), 
        tokenizer=AutoTokenizer.from_pretrained(model_path),
        framework="pt", device=0 if is_cuda else -1)

def split_ctrl_code(text):
    r = re.search(r'\[(?P<code>[a-z]+)\](?P<text>.+)', text)
    if r:
        return r.group("code").strip(), r.group("text").strip()
    return "", text
def remove_blanks(text):
    try:
        before, answers = text.split(SEP_TOK)
        answers = [x.strip() for x in answers.split(ANSWER_TOK)][:-1]
        answers = [x if x != EMPTY_TOK else '' for x in answers]
        for a in answers:
            if a == '':
                before = re.sub(r' %s' % re.escape(BLANK_TOK), a, before, count=1)
            else:
                before = re.sub(r'%s' % re.escape(BLANK_TOK), a, before, count=1)
        return before, answers
    except:
        return text, []

def batched_generate(generator, 
    examples, 
    temperature=1, 
    num_beams=None,
    num_return_sequences=3, 
    do_sample=True, 
    batch_size=128, **kwargs):
    preds_list = []
    with torch.no_grad():
        for e in (range(0, len(examples), batch_size)):
            preds_list += generator(
                examples[e:e+batch_size],
                temperature=temperature,
                return_tensors=True,
                num_beams=num_beams,
                max_length=1000,
                early_stopping=None if num_beams is None else True,
                do_sample=num_beams is None and do_sample,
                num_return_sequences=num_return_sequences, 
                **kwargs
            )
    return preds_list

def generate_on_prompts(generator, prompts, temperature=1, 
    num_beams=None, n=3, do_sample=True, batch_size=128, **kwargs):
    preds_list = batched_generate(generator, prompts,
        temperature=temperature, n=n, 
        num_beams=num_beams, 
        do_sample=do_sample, batch_size=batch_size, **kwargs)
    if len(prompts) == 1:
        preds_list = [preds_list]
    preds_list_cleaned = []
    for prompt, preds in zip(prompts, preds_list):
        prev_list = set()
        for s in preds:
            total_sequence = s["generated_text"].split(PERETURB_TOK)[-1]
            normalized, _ = remove_blanks(total_sequence)
            input_ctrl_code, normalized = split_ctrl_code(normalized)
            prev_list.add((input_ctrl_code, normalized))
        preds_list_cleaned.append(list(prev_list))
    return preds_list_cleaned

def get_prompts(doc, ctrl_codes, blanked_sents, is_complete_blank=True):
    prompts = []
    for tag, bt in itertools.product(ctrl_codes, blanked_sents):
        sep_tok = SEP_TOK if bt and is_complete_blank else ""
        prompts.append(f"{doc.text.strip()} {PERETURB_TOK} [{tag}] {bt.strip()} {sep_tok}".strip())
    return prompts

"""
def perturb_one_doc(generator, doc, tags, indexes, temperature=1, 
    num_beams=None, n=3, top_p=0.9, do_sample=True, batch_size=128):
    prompts = get_prompts(doc, tags, indexes)
    return generate_on_prompts(
        generator, prompts, 
        temperature=temperature,
        n=n, 
        top_p=top_p, 
        do_sample=do_sample,
        batch_size=batch_size
    )
"""


from typing import Callable, List, Dict, Tuple
from munch import Munch
from ..compute_perturbs import compute_edit_ops, SentenceMetadata
from ..helpers import unify_tags
import numpy as np
import itertools
from copy import deepcopy
def submodular_fn(candidate, chosen, distances):
    return np.mean(distances[candidate, chosen])

def greedy(submodular_fn, distances, k, chosen=[]):
    chosen = deepcopy(chosen)
    all_items = list(range(distances.shape[0]))
    current_value = 0
    z = 0
    while len(chosen) != k:
        best_gain = 0
        best_item = all_items[0]
        for i in all_items:
            gain = submodular_fn(i, chosen, distances)
            if gain > best_gain:
                best_gain = gain
                best_item = i
        chosen.append(best_item)
        all_items.remove(best_item)
        current_value += best_gain
        #print(z, current_value)
        z += 1
    return chosen

def similarity_on_metas(meta1: SentenceMetadata, meta2: SentenceMetadata):
    # if the cores are exactly the same:
    is_same_cores = meta1.primary.acore.lemma_ == meta2.primary.acore.lemma_ and \
        meta1.primary.bcore.lemma_ == meta2.primary.bcore.lemma_
    # change if the tag is the same
    is_same_tag = meta1.primary.tag == meta2.primary.tag
    is_same_dep = unify_tags(meta1.primary.dep) == unify_tags(meta2.primary.dep)
    if is_same_cores and is_same_tag: return 1
    if is_same_tag and is_same_dep: return 0.75
    if is_same_tag: return 0.5
    if is_same_dep: return 0.1
    return 0.0001

def select_diverse_perturbations(
    orig_and_perturb_pairs: List[Tuple[str, str]],
    nsamples: int,
    compute_sent_cosine_distance: Callable,
    process: Callable) -> List[Tuple[str, str]]:
    """Having each perturbation be represented by its token changes, 
    control code, and dependency tree strcuture, we greedily select the 
    ones that are least similar to those already selected. This tries 
    to avoid redundancy in common perturbations such as black -> white.

    Args:
        orig_and_perturb_pairs (List[Tuple[str, str]]):
            A list of (orig, perturb) text pair. Not necessarily the 
            all the same orig text.
        nsamples (int): Number of samples to select
        compute_sent_cosine_distance (Callable): Distance function.
        process (Callable): The spaCy processor.

    Returns:
        List[Tuple[str, str]]: A subsample of (orig, perturb) pairs.
    """
    texts = set([s for s in set(np.concatenate([list(p) for p in orig_and_perturb_pairs]))])
    doc_mapper = {text: process(text) for text in texts}
    metas = []
    for orig_text, perturb_text in orig_and_perturb_pairs:
        orig_doc, perturb_doc = doc_mapper[orig_text], doc_mapper[perturb_text]
        eops = compute_edit_ops(orig_doc, perturb_doc)
        meta = SentenceMetadata(eops)
        meta.compute_metadata(sentence_similarity=compute_sent_cosine_distance)
        if meta.primary:
            metas.append(meta)
    
    distances = np.zeros((len(metas), len(metas)))
    idxes = list(range(len(metas)))
    for i, j in itertools.combinations(idxes, 2):
        distance = 1 - similarity_on_metas(metas[i], metas[j])
        distances[i, j] = distance
        distances[j, i] = distance
    picked = greedy(submodular_fn, distances=distances, k=nsamples)
    return [(metas[i].primary.acore.doc.text, metas[i].primary.bcore.doc.text) for i in picked]

def extract_predict_label(raw_pred):
    raw_pred = sorted(raw_pred, key=lambda r: -r["score"])
    if raw_pred:
        return raw_pred[0]["label"]
    return None
PredFormat = List[Dict[str, float]]
def select_surprise_explanations(
    orig_text: str,
    perturb_texts: List[str],
    orig_pred: PredFormat, 
    perturb_preds: List[PredFormat], 
    feature_importance_dict: Dict[str, float], 
    processor: Callable, 
    agg_feature_importance_thred: float=0.1) -> List[Dict]:
    """Select surprising perturbations based on a feature importance map. 
    we estimate the expected change in prediction with feature attributions, 
    and select counterfactuals that violate these expectations, i.e., examples 
    where the real change in prediction is large even though importance scores 
    are low, and examples where the change is small but importance scores are high.

    Args:
        orig_doc (Doc): The original text.
        perturb_docs (List[Doc]): The perturbed texts.
        orig_pred (PredFormat): Prediction of the original text, in the format of 
            [{label: str, score: float}]
        perturb_preds (List[PredFormat]): Prediction of the perturbed texts,
            Each entry in the format of [{label: str, score: float}]
        feature_importance_dict (Dict[str, float]): The precomputed feature map.
        agg_feature_importance_thred (float, Optional): A threshold for classifying
            surprises.
        process (Callable): The spaCy processor.

    Returns:
        List[Dict]: Return the selected surprises in the format of:
        {
            case, [str]: Suprise filp | Suprise unflip, 
            pred, [str]: the prediction for the perturbation,
            changed_features, List[str]: the changed features
            perturb_doc: the perturbed doc))
        }
    """
    # compute whether the perturbation is a surprising one.
    def is_surprise(
        delta_pred_prob,
        is_pred_changed,
        agg_feature_importance,
        agg_feature_importance_thred=0.1):
        # is an important feature, but changing it only has a small effect
        if agg_feature_importance >= agg_feature_importance_thred and abs(agg_feature_importance) > abs(delta_pred_prob):
            return not is_pred_changed
        if agg_feature_importance <= agg_feature_importance_thred and abs(agg_feature_importance) < abs(delta_pred_prob):
            return is_pred_changed
        return False

    orig_doc = processor(orig_text)
    perturb_docs = [processor(perturb) for perturb in perturb_texts]
    orig_pred_dict = {p["label"]: p["score"] for p in orig_pred}
    orig_pred = extract_predict_label(orig_pred)
    # a dict that groups perturbations by the words changed
    perturbed_word_dict = {}
    for perturbs_pred, perturb_doc in zip(perturb_preds, perturb_docs):
        perturbs_pred_dict = {p["label"]: p["score"] for p in perturbs_pred}
        perturb_pred = extract_predict_label(perturbs_pred)
        eops = compute_edit_ops(orig_doc, perturb_doc, only_return_edit=True)
        # actual prediction change
        is_pred_changed = orig_pred != perturb_pred
        delta_pred_prob = abs(perturbs_pred_dict[orig_pred] - orig_pred_dict[orig_pred])
        # get the changed parts
        all_changed_features = []
        for eop in eops:
            try: tokens = eop.fromz_core if eop.op != "insert" else [eop.fromz_full.root]
            except: tokens = []
            all_changed_features += [t.text for t in tokens]
        agg_feature_importance=sum([feature_importance_dict[t] for t in all_changed_features])

        surprise = is_surprise(
                delta_pred_prob, is_pred_changed, 
                agg_feature_importance, agg_feature_importance_thred)
        for t in all_changed_features:
            if t not in perturbed_word_dict:
                perturbed_word_dict[t] = [] # initialize the list
            perturbed_word_dict[t].append(Munch(
                perturb_doc = perturb_doc,
                pred=perturb_pred,
                delta_pred_prob=delta_pred_prob,
                changed_features=all_changed_features,
                agg_feature_importance=agg_feature_importance,
                is_pred_changed=is_pred_changed,
                is_surprise=surprise,
                # equally spread the impact on delta to each of the changed feature
                weight=1/len(all_changed_features)
            ))

    selected_features = []
    # find the surprising flip and suripring unflip separately.
    for is_pred_changed in [True, False]:
        sorted_features = []
        # find abnormal tokens: token has small SHAP weights, but the delta is large on average; Or the other way around
        for feature, group in perturbed_word_dict.items():
            grouped_perturbations = [g for g in group if g.is_pred_changed==is_pred_changed]
            if not grouped_perturbations: continue
            feature_imp = feature_importance_dict[feature]
            deltas = [feature_imp] + [k.delta_pred_prob for k in grouped_perturbations]
            weights = [1] + [k.weight for k in grouped_perturbations]
            sorted_features.append((feature, feature_imp - np.average(deltas, weights=weights)))
        # always prioritize the ones with the largest difference
        sorted_features = sorted(sorted_features, key=lambda s: -abs(s[1]))
        for feature, _ in sorted_features:
            grouped_perturbations = [g for g in perturbed_word_dict[feature] if 
                                    g.is_pred_changed==is_pred_changed and g.is_surprise]
            if not grouped_perturbations: continue
            grouped_perturbations = sorted(
                grouped_perturbations, 
                key=lambda p: (
                    -1 * p.weight,
                    # if surprising flipp, select the ones with the least prediction change
                    #(1 if is_pred_changed else -1) *
                    #p.delta_pred_prob * p.weight
                    -1 * abs(p.delta_pred_prob - p.agg_feature_importance)
                ))
                    # the delta between the prediction change, and the aggregated SHAP weights of all changed tokens
                    #abs(p.delta_pred_prob - p.agg_feature_importance) ))
            selected_features.append(Munch(
                case=f"Suprise {'flip' if is_pred_changed else 'unflip'}", 
                pred=grouped_perturbations[0].pred,
                changed_features=grouped_perturbations[0].changed_features,
                perturb_text=grouped_perturbations[0].perturb_doc.text))
            break
    return selected_features


import scipy
from sentence_transformers import SentenceTransformer

#########################################################################
### compute similarity
#########################################################################
def load_distance_scorer(is_cuda):
    distance_scorer = SentenceTransformer('stsb-distilbert-base')
    if is_cuda: distance_scorer.to('cuda')
    return distance_scorer

def compute_sent_cosine_distance(s1, s2, similarity_scorer):
    s1 = [s1] if type(s1) == str else s1
    s2 = [s2] if type(s2) == str else s2
    s1_embed, s2_embed = similarity_scorer.encode(s1), similarity_scorer.encode(s2)
    distances = scipy.spatial.distance.cdist(s1_embed, s2_embed, "cosine")[0][0]
    # np.unravel_index(distances.argmin(), distances.shape)
    return distances



import numpy as np
from ..helpers import unify_tags, flatten_fillins
from .special_tokens import BLANK_TOK

def create_blanked_sents(doc, indexes=None):
    if indexes:
        if type(indexes[0]) == int: 
            indexes = [indexes]
        indexes_list = indexes #[indexes]
    else:
        indexes_list = get_random_idxes(
            doc, is_token_only=False, max_count=3)
    blanks = set([flatten_fillins(
        doc, indexes, [BLANK_TOK] * len(indexes)) \
        for indexes in indexes_list])
    return blanks

# the function for placing BLANKS.
def get_one_random_idx_set(
    doc, max_blank_block=3, req_dep=None, blank_type_prob=None, 
    pre_selected_idxes=None, is_token_only=False):
    if req_dep is not None:
        if type(req_dep) == str: req_dep = [req_dep]
        idx_range = [i for i, token in enumerate(doc) if token.dep_ in req_dep or unify_tags(token.dep_) in req_dep]
    else:
        idx_range = list(range(len(doc)))
    # only keep those pre_selected_idxes
    if pre_selected_idxes is not None:
        idx_range = [i for i in idx_range if i in pre_selected_idxes]
    max_blank_block = min(len(idx_range), max_blank_block)        
    #print(req_dep, idx_range)
    selected_indexes = []
    while max_blank_block > 0 and not selected_indexes:
        # if fixed the thing to change, then do one specific change
        n_perturb = np.random.choice(list(range(1, max_blank_block+1))) #if req_dep is None else 1
        replace_idx, total_run = -1, 1000
        while (total_run > 0 and n_perturb > 0): #and  len(span_and_edits) == 0:
            replace_idx = np.random.choice(idx_range)
            token = doc[replace_idx]
            if token.is_punct:
                total_run -= 1
                continue
            if blank_type_prob: p = blank_type_prob
            else:
                # if fixed the tree, then mostly use the tree
                if is_token_only:  p = [0.7, 0, 0.3]
                elif req_dep is None: p = [0.4, 0.35, 0.25]
                else: p = [0.1, 0.7, 0.2]
            is_replace_subtree = np.random.choice(["token", "subtree", "insert"], p=p)
            if is_replace_subtree == "subtree":
                start, end = token.left_edge.i, token.right_edge.i+1
            elif is_replace_subtree == "token":
                start, end = token.i, token.i+1
            else:
                start, end = token.i, token.i 
            if all([end < sstart or start > send for sstart, send in selected_indexes]):
                selected_indexes.append([start, end])
                n_perturb -= 1
            total_run -= 1
    return sorted(selected_indexes, key=lambda idx: (idx[0], idx[1]))


def get_random_idxes(doc, 
    pre_selected_idxes=None, 
    deps=None, is_token_only=False, 
    max_blank_block=3, max_count=None):
    unique_blanks = {str([[0, len(doc)]]): [[0, len(doc)]]}
    default_deps = [None, "", ["subj","obj"], ["aux", "ROOT"], ["conj", "modifier", "clause"]]
    if is_token_only: 
        unique_blanks = {}
    if deps is None: deps = default_deps
    for dep in deps:
        # for each different dep, get some blank
        rounds = 1 if dep is not None else 2
        if is_token_only:
            rounds = 5
        for _ in range(rounds):
            curr_idx = get_one_random_idx_set(
                doc, req_dep=dep, 
                max_blank_block=max_blank_block,
                pre_selected_idxes=pre_selected_idxes, 
                is_token_only=is_token_only) if dep != "" else None
            if curr_idx is not None:
                unique_blanks[str(curr_idx)] = curr_idx
    unique_blanks = list(unique_blanks.values())
    if max_count is not None:
        try:
            unique_blanks = list(np.random.choice(
                np.array(unique_blanks, dtype="object"), 
                min(len(unique_blanks), max_count), 
                replace=False))
        except:
            unique_blanks = unique_blanks[:max_count]
    return unique_blanks




import itertools
from spacy.tokens import Token, Span, Doc
import spacy
def _normalize_chunk(doc, start_idx, end_idx):
    # end idx is not included: [start, end)
    # the function is for normalizing the chunk so it does not start/end with punctuation
    punctuation = "''\"!, -.:;<=>?\^_|~”’ "
    if not doc:
        return None
    end_idx = min([len(doc), end_idx])
    start_idx = max([0, start_idx])
    while start_idx < end_idx and (doc[start_idx].text in punctuation or doc[start_idx].is_punct or doc[start_idx].is_space):
        start_idx += 1
    while start_idx < end_idx and (doc[end_idx-1].text in punctuation or doc[end_idx-1].is_punct or doc[end_idx-1].is_space):
        end_idx -= 1
    return doc[start_idx:end_idx]


def _normalize_span(span):
    #print(span,type(span))
    if type(span) == Span: return _normalize_chunk(span.doc, span.start, span.end)
    if type(span) == Token: return _normalize_chunk(span.doc, span.i, span.i+1)
    return None

def unify_tags(dep):
    norm_dep = dep
    if dep.endswith("mod") or dep in ["poss", "appos", "compound", "meta", "nn", "neg"]:
        # "nummod", "quantmod"
        norm_dep = "modifier"
    elif dep.endswith("cl") or dep.endswith("comp") or dep in ["mark", "parataxis", "attr", "oprd", "prep", "cop"]:
        norm_dep = "clause"
    elif dep.endswith("obj") or dep in ["dative", "obl"]:
        norm_dep = "obj"
    elif "subj" in dep:
        norm_dep = "subj"
    elif dep in ["cc", "conj", "preconj", "predet"]:
        norm_dep = "conj" 
    elif dep == "ROOT": norm_dep = "ROOT"
    elif dep == "aux": norm_dep = "aux"
    else: norm_dep = "others"
    return norm_dep


def flatten_fillins(doc, indxes, fillins_by_idxes, is_return_text=True):
    """Use index and corresponding fill ins to create a modified string.
    Useful for creating blanks.

    Args:
        doc (Doc): spacy doc
        indxes ([int, int][]): the indexes where the text should be modified
        fillins_by_idxes (str[][]): a list of candidate fillins at each space.
        is_return_text (bool, optional): If true, return the first possible 
        changed text. Otherwise return a list. Defaults to False.

    Returns:
        [type]: [description]
    """
    text_arr = []
    prev = 0
    zipped = list(zip(indxes, fillins_by_idxes))
    zipped = sorted(zipped, key=lambda z: z[0][0])
    for (start, end), fillins in zipped:
        if type(fillins) == str:
            fillins = [fillins]
        space = " " if end == 0 or doc[end-1].text_with_ws.endswith(' ') else ""
        text_arr.append([doc[prev:start].text_with_ws])
        text_arr.append([f.strip() + space for f in fillins])
        prev = end
    #if len(doc) < prev:
    text_arr.append([doc[prev:].text])
    otexts = ["".join(s) for s in itertools.product(*text_arr)]
    if is_return_text:
        return otexts[0].strip() if otexts else ""
    else:
        return otexts.strip()


class _WhitespaceSpacyTokenizer:
    """
    Spacy doesn't assume that text is tokenised. Sometimes this
    is annoying, like when you have gold data which is pre-tokenised,
    but Spacy's tokenisation doesn't match the gold. This can be used
    as follows:
    nlp = spacy.load("en_core_web_md")
    # hack to replace tokenizer with a whitespace tokenizer
    nlp.tokenizer = _WhitespaceSpacyTokenizer(nlp.vocab)
    ... use nlp("here is some text") as normal.
    """
    def __init__(self, vocab):
        self.vocab = vocab
    def __call__(self, text):
        words = text.strip().split(" ")
        spaces = [True] * len(words)
        return Doc(self.vocab, words=words, spaces=spaces)

def create_processor(is_space_tokenizer):
    nlp = spacy.load("en_core_web_sm")
    if is_space_tokenizer:
        nlp.tokenizer = _WhitespaceSpacyTokenizer(nlp.vocab)
    return nlp